package com.sillycycle.bagleyd.cubes;

/*
 * @(#)CubesCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>CubesCanvas</code> class does all the graphics for
 * <code>CubesFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
*/

public class CubesCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 256;
	static final int MAX_SLICES = 16;

	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	static final int INWARDS = 4;
	static final int OUTWARDS = 5;
	static final int COORD = 6;
	static final int BLOCKED = -3;
	static final int SPACE = -2;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "cubes.dat";
	static final String SYMBOL = ":";
	static final String PICTUREFILE = CubesInterface.IMAGE;
	static final int PICTUREX = 512;
	static final int PICTUREY = 512;

	Color       foreground, background;
	Color       blockColor;
	int[]       blockOfPosition = null;
	int[]       startPosition = null;
	int	 spacePosition, startSpace;
	int[]       currentRow = new int[3];
	int	 currentPosition = -1, currentSpace = -1;
	int	 currentDirection, lastDirection;
	boolean     started, cheat, vertical;
	int	 base;
	int	 delay, numSlices;
	transient CubesSize  blocks = new CubesSize();
	int	 sizeRect, sizeBlock;
	Point       coreSize;
	Point       offset;
	Point       blockSize, faceSize, puzzleSize;
	Point       delta, puzzleOffset;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     firstPaint = false, resizePaint = false;
	boolean     framePaint = false, allBlocksPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false;
	boolean     focus = true;
	boolean     number = CubesInterface.DEFAULT_NUMBER;
	Random      generator;
	transient CubesStack  undo = null, redo = null;
	transient CubesSolve  solve = null;
	String      stringSave = null;
	String[]    token = null;
	transient Image   bufferedFrameImage = null;
	transient Image[]   bufferedStartImage = new Image[2];
	transient Graphics   bufferedFrameGraphics = null;
	transient Graphics[]  bufferedStartGraphics = new Graphics[2];

	@Override
	protected String paramString() {
		return super.paramString() +
		",blocksX=" + blocks.x +
		",blocksY=" + blocks.y +
		",blocksZ=" + blocks.z +
		",base=" + base +
		",blockColor=" + blockColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkBlocks() {
		if (blocks.x < CubesInterface.MIN_BLOCKS) {
			System.out.println(blocks.x +
			" number of cubes in a row incorrect, must be at least " +
			CubesInterface.MIN_BLOCKS +
			", defaulting to " +
			CubesInterface.DEFAULT_BLOCKSX);
			blocks.x = CubesInterface.DEFAULT_BLOCKSX;
		}
		if (blocks.y < CubesInterface.MIN_BLOCKS) {
			System.out.println(blocks.y +
			" number of cubes in a row incorrect, must be at least " +
			CubesInterface.MIN_BLOCKS +
			", defaulting to " +
			CubesInterface.DEFAULT_BLOCKSY);
			blocks.y = CubesInterface.DEFAULT_BLOCKSY;
		}
		if (blocks.z < CubesInterface.MIN_BLOCKS) {
			System.out.println(blocks.z +
			" number of cubes in a row incorrect, must be at least " +
			CubesInterface.MIN_BLOCKS +
			", defaulting to " +
			CubesInterface.DEFAULT_BLOCKSZ);
			blocks.z = CubesInterface.DEFAULT_BLOCKSZ;
		}
		if (base < CubesInterface.MIN_BASE ||
				base > CubesInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			CubesInterface.MIN_BASE +
			" and " +
			CubesInterface.MAX_BASE +
			" inclusive, defaulting to " +
			CubesInterface.DEFAULT_BASE);
			base = CubesInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	int toPosition(int column, int row, int stack) {
		return (column + blocks.x * row + sizeRect * stack);
	}

	int typeOfMix() {
		if (blocks.x >= blocks.z && blocks.y >= blocks.z) {
			return 1;
		} else if (blocks.x >= blocks.y && blocks.z >= blocks.y) {
			return 2;
		} else if (blocks.y >= blocks.x && blocks.z >= blocks.x) {
			return 3;
		}
		return -1;
	}

	int toInvPosition(int column, int row, int stack) {
		int mix = typeOfMix();

		switch (mix) {
		case 1: // swap x/y
			return (row + blocks.y * column + sizeRect * stack);
		case 2: // swap x/z
		case 3: // swap x/z again
			return (blocks.z * column + blocks.x * blocks.z * row + stack);
		default:
			return -1;
		}
	}

	boolean decideSwap() {
		int mix = typeOfMix();

		switch (mix) {
		case 1: // swap x/y
			return (((blocks.x >> 1) & 1) == 1
				&& ((blocks.y >> 1) & 1) == 1
				&& (blocks.z & 1) == 1);
		case 2: // swap x/z
			return (((blocks.x >> 1) & 1) == 1
				&& ((blocks.z >> 1) & 1) == 1
				&& (blocks.y & 1) == 1);
		case 3: // swap y/z
			return (((blocks.y >> 1) & 1) == 1
				&& ((blocks.z >> 1) & 1) == 1
				&& (blocks.x & 1) == 1);
		default:
			return false;
		}
	}

	/*public static void reformatPuzzle(CubesSize blocks, int vector) {
		if (vector <= 0) {
			if (blocks.x < blocks.y)
				blocks.y = blocks.x;
			if (blocks.y < blocks.z)
				blocks.z = blocks.y;
		} else if (vector > 0) {
			if (blocks.z > blocks.y)
				blocks.y = blocks.z;
			if (blocks.y > blocks.x)
				blocks.x = blocks.y;
		}
	}*/

	int toColumn(int pos) {
		return ((pos % sizeRect) % blocks.x);
	}

	int toRow(int pos) {
		return ((pos % sizeRect) / blocks.x);
	}

	int toStack(int pos) {
		return (pos / sizeRect);
	}

	int cartesianX(int pos) {
		int x;

		x = toColumn(pos) * offset.x + delta.x + puzzleOffset.x + 1;
		if (!vertical) {
			x += toStack(pos) * (blocks.x * offset.x + 5);
		}
		return x;
	}

	int cartesianY(int pos) {
		int y;

		y = toRow(pos) * offset.y + delta.y + puzzleOffset.y + 1;
		if (vertical) {
			y += toStack(pos) * (blocks.y * offset.y + 5);
		}
		return y;
	}

	int blockNFromSpace(int n, int direction) {
		int pos = spacePosition;

		switch(direction) {
		case LEFT:
		case RIGHT:
			return (pos + ((direction == RIGHT) ? -n : n));
		case TOP:
		case BOTTOM:
			return (pos + (blocks.x * ((direction == BOTTOM) ?
				-n : n)));
		default:
			return (pos + (sizeRect * ((direction == OUTWARDS) ?
				-n : n)));
		}
	}

	int solvedPosition(int block) {
		return (block + 1) % sizeBlock;
	}

	public boolean checkSolved() {
		int i;

		for (i = 1; i < sizeBlock; i++)
			if (blockOfPosition[i - 1] != i)
				return false;
		started = false;
		return true;
	}

	public static void fill3DRect(Graphics g, int x, int y, int width, int height,
			boolean raised) {
		Color c = g.getColor();
		Color brighter = c.brighter();
		Color darker = c.darker();

		if (!raised) {
			g.setColor(darker);
		}
		if (width > 2 && height > 2)
			g.fillRect(x + 1, y + 1, width - 2, height - 2);
		g.setColor(raised ? brighter : darker);
		if (height > 1)
			g.drawLine(x, y, x, y + height - 1);
		if (width > 2)
			g.drawLine(x + 1, y, x + width - 2, y);
		if (height > 1)
			g.drawLine(x + 1, y, x + 1, y + height - 1);
		if (width > 1)
			g.drawLine(x + 2, y + 1, x + width - 1, y + 1);
		g.setColor(raised ? darker : c);
		if (width > 1 && height > 1)
			g.drawLine(x + 1, y + height - 1,
				x + width - 1, y + height - 1);
		if (width > 1 && height > 2)
			g.drawLine(x + width - 1, y,
				x + width - 1, y + height - 2);
		if (width > 2 && height > 2)
			g.drawLine(x + 2, y + height - 2,
				x + width - 2, y + height - 2);
		if (width > 2 && height > 3)
			g.drawLine(x + width - 2, y + 1,
				x + width - 2, y + height - 3);
		g.setColor(c);
	}

	static void drawShadow(Graphics g, int startX, int startY,
			int width, int height) {
		g.drawLine(startX, startY, startX + width, startY);
		g.drawLine(startX, startY, startX, startY + height);
	}

	void copyImage(Graphics g, Image i,
			int dX, int dY, int sX, int sY,
			int width, int height) {
		if (number) {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX, sY, sX + width, sY + height, this);
		} else {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX * PICTUREX / coreSize.x,
				sY * PICTUREY / coreSize.y,
				(sX + width) * PICTUREX / coreSize.x,
				(sY + height) * PICTUREY / coreSize.y, this);
		}
	}

	void drawBlock(Graphics g, int pos,
			boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		int dx, dy, sx, sy;
		int block = (blockOfPosition[pos] + sizeBlock - 1) %
			sizeBlock;

		dx = cartesianX(pos) + offsetX + pressedOffset;
		dy = cartesianY(pos) + offsetY + pressedOffset;
		sx = cartesianX(block);
		sy = cartesianY(block);
		if (blank) {
			g.setColor((erase) ? background : blockColor);
			g.fillRect(dx, dy, blockSize.x, blockSize.y);
			return;
		}
		copyImage(g, bufferedStartImage[pressedOffset],
			dx, dy,
			sx, sy,
			blockSize.x, blockSize.y);
	}

	void drawBlock(int pos, boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawBlock(g, pos, blank, erase,
				pressedOffset, offsetX, offsetY);
		} finally {
			g.dispose();
		}
	}

	void drawBufferedBlock(Graphics g, int pos, int pressedOffset) {
		int dx, dy;
		Color fillColor, textColor;
		int i, block = solvedPosition(pos);
		int digitX = 0, digitY, fontSize;

		dx = cartesianX(pos) + pressedOffset;
		dy = cartesianY(pos) + pressedOffset;
		fillColor = blockColor;
		textColor = background.darker();
		if (pressedOffset != 0) {
			g.setColor(fillColor.darker());
			drawShadow(g,
				dx - pressedOffset, dy - pressedOffset,
				blockSize.x - 1, blockSize.y - 1);
		}
		g.setColor(fillColor);
		fill3DRect(g, dx, dy, blockSize.x, blockSize.y,
			pressedOffset == 0);
		g.setColor(textColor);
		fontSize = java.lang.Math.min(blockSize.x,
			blockSize.y) / 2;
		g.setFont(new Font("arial", Font.PLAIN,
			fontSize));
		i = block;
		while (i >= 1) {
			i /= base;
				digitX += -(fontSize / 3);
		}
		digitY = fontSize / 3;
		g.drawString(Integer.toString(block, base).toUpperCase(),
			dx + blockSize.x / 2 + digitX +
			blockSize.x / 25,
			dy + blockSize.y / 2 + digitY);
	}

	void drawAllBlocks() {
		int	k;

		for (k = 0; k < sizeBlock; k++)
			drawBlock(k, (blockOfPosition[k] <= 0),
				(blockOfPosition[k] <= 0), 0, 0, 0);
	}

	void drawAllBufferedBlocks() {
		if (number) {
			int k, l;

			for (k = 0; k < sizeBlock; k++) {
				for (l = 0; l < 2; l++) {
					drawBufferedBlock(
						bufferedStartGraphics[l],
						k, l);
				}
			}
		} else {
			URL url = CubesFrame.class.getResource("/" + PICTUREFILE);

			if (url == null)
				bufferedStartImage[1] = null;
			else
				bufferedStartImage[1] = Toolkit.getDefaultToolkit().getImage(url);
			bufferedStartImage[0] = bufferedStartImage[1];
		}

	}

	void drawBuffer(Graphics g, Image i) {
		g.drawImage(i, 0, 0, coreSize.x, coreSize.y, this);
	}

	void drawBuffer(Image i) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawBuffer(g, i);
		} finally {
			g.dispose();
		}
	}

	void animateSlide(int numOfBlocks, int dir,
			int fast, boolean logMoves) {
		Graphics g = getGraphics();
		int inc, aBlock;
		int fillBeginPos, fillBeginI = 0, gapI = 0, moveI = 0;
		int dx, dy, pos, posNext;
		int ix = 0, iy = 0;
		int numBlocks = numOfBlocks;

		if (g == null)
			return;
		try {
			fillBeginPos = blockNFromSpace(numBlocks, dir);
			if ((dir == RIGHT) || (dir == LEFT)) {
				fillBeginI = cartesianX(fillBeginPos);
				gapI = blockSize.x * fast / numSlices;
				moveI = blockSize.x + delta.x;
			} else if ((dir == TOP) || (dir == BOTTOM)) {
				fillBeginI = cartesianY(fillBeginPos);
				gapI = blockSize.y * fast / numSlices;
				moveI = blockSize.y + delta.y;
			}
			if (gapI == 0)
				gapI++;
			if (numBlocks < 0)
				numBlocks = -numBlocks;
			for (inc = 0; inc < moveI + gapI; inc += gapI) {
			  if (inc > moveI)
			    inc = moveI;
			  for (aBlock = 0; aBlock < numBlocks; aBlock++) {
			    posNext = blockNFromSpace(aBlock + 1, dir);
			    if (logMoves && inc == 0) {
			      ((CubesFrame) frame).callback(
				CubesInterface.ACTION_MOVED);
			      undo.setMove(dir);
			      redo.flushMoves();
			    }
			    /* Calculate deltas */
			    dx = cartesianX(posNext);
			    dy = cartesianY(posNext);
			    if ((dir == RIGHT) || (dir == LEFT)) {
			      ix = ((dir == RIGHT) ? inc : -inc);
			      iy = 0;
			    } else if ((dir == TOP) || (dir == BOTTOM)) {
			      ix = 0;
			      iy = ((dir == BOTTOM) ? inc : -inc);
			    }
			    drawBlock(posNext, false, false, 0, ix, iy);
			    /* Erase old slivers */
			    ix += dx;
			    iy += dy;
			    g.setColor(background);
			    if (aBlock < numBlocks - 1) {
			      switch (dir) {
			      case TOP:
				g.fillRect(ix, iy + blockSize.y,
				  blockSize.x, gapI);
				break;
			      case LEFT:
				g.fillRect(ix + blockSize.x, iy,
				  gapI, blockSize.y);
				break;
			      case RIGHT:
				g.fillRect(ix - gapI, iy, gapI, blockSize.y);
				break;
			      case BOTTOM:
				g.fillRect(ix, iy - gapI, blockSize.x, gapI);
				break;
			      default:
			      }
			    } else {
			      switch (dir) {
			      case TOP:
				g.fillRect(ix, iy + blockSize.y,
				  blockSize.x, fillBeginI - iy);
				break;
			      case RIGHT:
				g.fillRect(fillBeginI, iy,
				  ix - fillBeginI, blockSize.y);
				break;
			      case BOTTOM:
				g.fillRect(ix, fillBeginI,
				  blockSize.x, iy - fillBeginI);
				break;
			      case LEFT:
				g.fillRect(ix + blockSize.x, iy,
				  fillBeginI - ix, blockSize.y);
				break;
			      default:
			      }
			    }
			  }
			  try {
			    Thread.sleep(delay / fast);
			  } catch (InterruptedException e) {
				  // ignore
			  }
			}
			pos = spacePosition;
			for (aBlock = 0; aBlock < numBlocks; aBlock++) {
			  posNext = blockNFromSpace(aBlock + 1, dir);
			  blockOfPosition[pos] = blockOfPosition[posNext];
			  pos = posNext;
			}
			spacePosition = fillBeginPos;
			blockOfPosition[fillBeginPos] = 0;
		} finally {
			g.dispose();
		}
	}

	void animateMove(int fromPos, int fast) {
		Graphics g = getGraphics();
		if (g == null)
			return;
		try {
			int pos = spacePosition;
			int dx, dy, fx, fy, sx, sy;
			int xToCenter, yToCenter;
			int incTo, incAt;
			int block = (blockOfPosition[pos] + sizeBlock - 1) %
				sizeBlock;

			dx = cartesianX(pos);
			dy = cartesianY(pos);
			fx = cartesianX(fromPos);
			fy = cartesianY(fromPos);
			sx = cartesianX(block) - 1;
			sy = cartesianY(block) - 1;

			xToCenter = dx + blockSize.x / 2;
			yToCenter = dy + blockSize.y / 2;
			incTo = java.lang.Math.min(blockSize.x,
				blockSize.y) / 2;
			incAt = java.lang.Math.max(incTo / (2 * numSlices), 1);

			g.setColor(background);
			g.fillRect(fx, fy, incAt, blockSize.y);
			g.fillRect(fx, fy, blockSize.x, incAt);
			g.fillRect(fx + blockSize.x - incAt, fy,
				incAt, blockSize.y);
			g.fillRect(fx, fy + blockSize.y - incAt,
				blockSize.x, incAt);
			for (int inc = incAt; inc <= incTo; inc += incAt) {
				//draw disappear block
				if (blockSize.x - 2 * inc > 0 &&
						blockSize.y - 2 * inc > 0) {
					g.setColor(background);
					g.fillRect(fx, fy,
						inc, blockSize.y);
					g.fillRect(fx, fy,
						blockSize.x, inc);
					g.fillRect(fx + blockSize.x - inc, fy,
						inc, blockSize.y);
					g.fillRect(fx, fy + blockSize.y - inc,
						blockSize.x, inc);
					g.drawImage(bufferedStartImage[0],
						fx + inc, fy + inc,
						fx + blockSize.x - inc,
						fy + blockSize.y - inc,
						sx - 1, sy - 1,
						sx + blockSize.x + 1,
						sy + blockSize.y + 1, this);
				}
				//draw appear block
				if (blockSize.x > blockSize.y) {
		g.drawImage(bufferedStartImage[0],
			xToCenter - inc - (blockSize.x / 2 - blockSize.y / 2),
			yToCenter - inc,
			xToCenter + inc + (blockSize.x / 2 - blockSize.y / 2),
			yToCenter + inc,
			sx - 1, sy - 1,
			sx + blockSize.x + 1, sy + blockSize.y + 1, this);
				} else {
		g.drawImage(bufferedStartImage[0],
			xToCenter - inc,
			yToCenter - inc - (blockSize.y / 2 - blockSize.x / 2),
			xToCenter + inc,
			yToCenter + inc + (blockSize.y / 2 - blockSize.x / 2),
			sx - 1, sy - 1,
			sx + blockSize.x + 1, sy + blockSize.y + 1, this);
				}
				try {
					Thread.sleep(delay / fast);
				} catch (InterruptedException e) {
					// ignore
				}
			}
			g.setColor(background);
			g.fillRect(fx, fy, blockSize.x, blockSize.y);
			drawBlock(pos, false, false, 0, 0, 0);
			spacePosition = fromPos;
		} finally {
			g.dispose();
		}
	}

	void resetBlocks(/*int vector*/) {
		int	i, j, switch1 = 0, switch2 = 0;

		solve.setFlag(false);
		//reformatPuzzle(blocks, vector);
		sizeRect = blocks.x * blocks.y;
		sizeBlock = sizeRect * blocks.z;
		blockOfPosition = new int[sizeBlock];
		startPosition = new int[sizeBlock];
		/*spacePosition = sizeBlock - 1;
		blockOfPosition[sizeBlock - 1] = 0;
		for (i = 1; i < sizeBlock; i++)
			blockOfPosition[i - 1] = i;*/
		/* makes puzzle too trivial
		for (i = 0; i < sizeBlock; i++)
			blockOfPosition[i] = sizeBlock - i;
		spacePosition = 0;
		blockOfPosition[0] = 0;*/
		for (i = 0; i < sizeBlock; i++)  {
			j = toInvPosition(toColumn(i), toRow(i), toStack(i)) + 1;
			blockOfPosition[i] = j;
			if (j == sizeBlock - 2)
				switch1 = i;
			else if (j == sizeBlock - 1)
				switch2 = i;
		}
		spacePosition = sizeBlock - 1;
		blockOfPosition[sizeBlock - 1] = 0;
		if (decideSwap()) {
			i = blockOfPosition[switch1];
			blockOfPosition[switch1] = blockOfPosition[switch2];
			blockOfPosition[switch2] = i;
		}
		if ((blocks.x == 2 && blocks.y == 2 && blocks.z == 1) ||
				(blocks.x == 2 && blocks.y == 1 && blocks.z == 2) ||
				(blocks.x == 1 && blocks.y == 2 && blocks.z == 2)) {
			spacePosition = 0;
			blockOfPosition[0] = 0;
			blockOfPosition[1] = 3;
			blockOfPosition[2] = 2;
			blockOfPosition[3] = 1;
		}
		if ((blocks.x > 1 && blocks.y == 1 && blocks.z == 1) ||
				(blocks.x == 1 && blocks.y > 1 && blocks.z == 1) ||
				(blocks.x == 1 && blocks.y == 1 && blocks.z > 1)) {
			int max = (blocks.x > blocks.y) ? blocks.x :
				(blocks.y > blocks.z) ? blocks.y : blocks.z;

			spacePosition = 0;
			blockOfPosition[0] = 0;
			for (i = 1; i < max; i++)
				blockOfPosition[i] = i;
		}
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = puzzleOffset.x;
		int offsetY = puzzleOffset.y;

		if (g == null)
			return;
		if (show) {
			int sumX, sumY, sumZ, dz;

			g.setColor(background.darker().darker());
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY,
				coreSize.x, offsetY);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX, 0,
				offsetX, coreSize.y);
			if (focus)
				g.setColor(foreground);
			else
				g.setColor(background);
			sumX = blocks.x * offset.x + 4;
			sumY = blocks.y * offset.y + 4;
			if (vertical) {
				sumZ = offsetY;
				dz = sumY;
			} else {
				sumZ = offsetX;
				dz = sumX;
			}
			for (int k = 0; k < blocks.z; k++) {
				if (vertical) {
					if (k > 0)
						g.fillRect(offsetX, sumZ + k - 1,
							sumX, 1);
					draw3DFrame(focus,
						offsetX, sumZ + k, sumX, sumY);
				} else {
					if (k > 0)
						g.fillRect(sumZ + k - 1, offsetY,
							1, sumY);
					draw3DFrame(focus,
						sumZ + k, offsetY, sumX, sumY);
				}
				sumZ += dz;
			}
		} else {
			/*g.setColor(Color.white);
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY - 1,
				coreSize.x, offsetY);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX - 1, 0,
				offsetX, coreSize.y);*/
			g.setColor(background);
			g.fillRect(offsetX + 1, offsetY + 1,
				coreSize.x - 2 - 2 * offsetX,
				coreSize.y - 2 - 2 * offsetY);
		}
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void moveNoBlocks() {
		((CubesFrame) frame).callback(CubesInterface.ACTION_IGNORE);
	}

	void moveBlocks(int from, int fast) {
		int	tempBlock;

		tempBlock = blockOfPosition[from];
		if (fast != INSTANT && delay > 0) {
			int i = currentRow[0] = toColumn(from);
			int j = currentRow[1] = toRow(from);
			int k = currentRow[2] = toStack(from);
			int l = i + blocks.x * j - spacePosition % sizeRect;
			/* int m = i + blocks.x * j + sizeRect * k - spacePosition; */
			/* Order important if blocks.x = 1 */
			if (l % blocks.x == 0 && k == toStack(spacePosition)) {
				// vertical moves
				if (l < 0) {
					animateSlide(-l / blocks.x, BOTTOM,
						fast, false);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				} else if (l > 0) {
					animateSlide(l / blocks.x, TOP,
						fast, false);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				}
				return;
			} else if (l / blocks.x == 0 &&
				j == toRow(spacePosition) &&
				k == toStack(spacePosition)) {
				if (l < 0) {
					animateSlide(-l, RIGHT, fast, false);
				} else {	/* (l > 0) */
					animateSlide(l, LEFT, fast, false);
				}
				if (((CubesFrame) frame).getToggleSound())
					((CubesFrame) frame).playBumpAudio();
				return;
			}
			blockOfPosition[from] =
				blockOfPosition[spacePosition];
			blockOfPosition[spacePosition] = tempBlock;
			animateMove(from, fast);
			if (((CubesFrame) frame).getToggleSound())
				((CubesFrame) frame).playDripAudio();
		} else {
			blockOfPosition[from] =
				blockOfPosition[spacePosition];
			blockOfPosition[spacePosition] = tempBlock;
			drawBlock(spacePosition, false, false, 0, 0, 0);
			tempBlock = spacePosition - from;
			spacePosition = from;
			drawBlock(from, true, true, 0, 0, 0);
			if (((CubesFrame) frame).getToggleSound()) {
				if (java.lang.Math.abs(tempBlock) >= sizeRect)
					((CubesFrame) frame).playDripAudio();
				else
					((CubesFrame) frame).playBumpAudio();
			}
		}
	}

	boolean checkMoveBlocksDir(int direction) {
		switch (direction) {
		case TOP:
			if (toRow(spacePosition) < blocks.y - 1) {
				currentPosition = spacePosition + blocks.x;
				return true;
			}
			break;
		case RIGHT:
			if (toColumn(spacePosition) > 0) {
				currentPosition = spacePosition - 1;
				return true;
			}
			break;
		case BOTTOM:
			if (toRow(spacePosition) > 0) {
				currentPosition = spacePosition - blocks.x;
				return true;
			}
			break;
		case LEFT:
			if (toColumn(spacePosition) < blocks.x - 1) {
				currentPosition = spacePosition + 1;
				return true;
			}
			break;
		case INWARDS:
			if (toStack(spacePosition) < blocks.z - 1) {
				currentPosition = spacePosition + sizeRect;
				return true;
			}
			break;
		case OUTWARDS:
			if (toStack(spacePosition) > 0) {
				currentPosition = spacePosition - sizeRect;
				return true;
			}
			break;
		default:
			{
				System.out.println("checkMoveBlocksDir: direction"
					+ direction);
			}
		}
		return false;
	}

	boolean movePuzzleDir(int direction) {
		if (checkMoveBlocksDir(direction)) {
			currentDirection = direction;
			movePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	void movePuzzle(int direction, boolean control) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (control) {
			switch (direction) {
			case TOP:
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_INCY);
				break;
			case RIGHT:
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_INCX);
				break;
			case BOTTOM:
				if (blocks.y <= CubesInterface.MIN_BLOCKS)
					return;
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_DECY);
				break;
			case LEFT:
				if (blocks.x <= CubesInterface.MIN_BLOCKS)
					return;
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_DECX);
				break;
			case INWARDS:
				if (blocks.z <= CubesInterface.MIN_BLOCKS)
					return;
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_DECZ);
				break;
			case OUTWARDS:
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_INCZ);
				break;
			default:
				System.out.println("movePuzzle: should not get here");
			}
			return;
		}
		if (checkSolved()) {
			moveNoBlocks();
			return;
		}
		if (!movePuzzleDir(direction)) {
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_BLOCKED);
		}
	}

	void movePuzzleDelay(int direction) {
		movePuzzleDir(direction);
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	boolean exchangeBlocks(int pos1, int pos2) {
		int	tempBlock;

		if ((blockOfPosition[pos1] <= 0) || (blockOfPosition[pos2] <= 0))
			return false;
		tempBlock = blockOfPosition[pos1];
		blockOfPosition[pos1] = blockOfPosition[pos2];
		blockOfPosition[pos2] = tempBlock;
		return true;
	}

	void discreteMoves(int numBlocks, int dir) {
		moveBlocks(numBlocks, NORMAL);
		((CubesFrame) frame).callback(
			CubesInterface.ACTION_MOVED);
		undo.setMove(dir);
		redo.flushMoves();
	}

	int positionToBlock(int x, int y) {
		int i, j, k;
		if (vertical) {
			i = (x - delta.x / 2 -
				puzzleOffset.x) / offset.x;
			j = ((y - delta.y / 2 -
				puzzleOffset.y) % (blocks.y * offset.y +
				delta.y + 4)) / offset.y;
			k = (y - delta.y / 2 -
				puzzleOffset.y) / (blocks.y * offset.y +
				delta.y + 4);
		}
		else {
			i = ((x - delta.x / 2 -
				puzzleOffset.x) % (blocks.x * offset.x +
				delta.x + 4)) / offset.x;
			j = (y - delta.y / 2 -
				puzzleOffset.y) / offset.y;
			k = (x - delta.x / 2 -
				puzzleOffset.x) / (blocks.x * offset.x +
				delta.x + 4);
		}
		if (i >= 0 && j >= 0 && k >= 0 &&
			i < blocks.x && j < blocks.y && k < blocks.z)
			return toPosition(i, j, k);
		return -1;
	}

	int movableBlock() {
		int i, j, k, l, m;

		i = currentRow[0];
		j = currentRow[1];
		k = currentRow[2];
		l = i + blocks.x * j - spacePosition % sizeRect;
		m = i + blocks.x * j + sizeRect * k - spacePosition;

		/* Order important if blocks.x = 1 */
		if (l % blocks.x == 0 && k == toStack(spacePosition)) {
			/* (Clicked on a space?) */
			if (l == 0) {
				return SPACE;
			}
		} else if (!(l / blocks.x == 0 &&
				j == toRow(spacePosition) &&
				k == toStack(spacePosition)) &&
				(m % sizeRect != 0)) {
			return BLOCKED;
		}
		return 0;
	}

	void selectBlocks() {
		int	l, m, n;
		int	i = currentRow[0];
		int	j = currentRow[1];
		int	k = currentRow[2];

		l = i + blocks.x * j - (spacePosition % sizeRect);
		m = i + blocks.x * j + sizeRect * k - spacePosition;
		/* Order important if blocks.x = 1 */
		if (l % blocks.x == 0 && k == toStack(spacePosition)) {
			// vertical moves
			if (l < 0) {
				if (delay > 0) {
					animateSlide(-l / blocks.x, BOTTOM,
						NORMAL, true);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				} else {
					for (n = 1; n <= -l / blocks.x; n++)
						discreteMoves(spacePosition - blocks.x,
							BOTTOM);
				}
				started = true;
			} else if (l > 0) {
				if (delay > 0) {
					animateSlide(l / blocks.x, TOP,
						NORMAL, true);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				} else {
					for (n = 1; n <= l / blocks.x; n++)
						discreteMoves(spacePosition + blocks.x,
							TOP);
				}
				started = true;
			} else {	/* (l == 0) */
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_SPACE);
				return;
			}
		} else if (l / blocks.x == 0 &&
			j == toRow(spacePosition) &&
			k == toStack(spacePosition)) {
			if (l < 0) {
				if (delay > 0) {
					animateSlide(-l, RIGHT, NORMAL, true);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				} else {
					for (n = 1; n <= -l % blocks.x; n++)
						discreteMoves(spacePosition - 1,
							RIGHT);
				}
				started = true;
			} else {	/* (l > 0) */
				if (delay > 0) {
					animateSlide(l, LEFT, NORMAL, true);
					if (((CubesFrame) frame).getToggleSound())
						((CubesFrame) frame).playBumpAudio();
				} else {
					for (n = 1; n <= l % blocks.x; n++)
						discreteMoves(spacePosition + 1,
							LEFT);
				}
				started = true;
			}
		} else if (m % sizeRect == 0) {
			if (m < 0) {
				for (n = 1; n <= -toStack(m); n++) {
					discreteMoves(spacePosition - sizeRect,
						OUTWARDS);
				}
			} else {	/* (m > 0) */
				for (n = 1; n <= toStack(m); n++) {
					discreteMoves(spacePosition + sizeRect,
						INWARDS);
				}
			}
		} else {
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_BLOCKED);
			return;
		}
		if (checkSolved())
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_SOLVED);
	}

	void randomizeBlocks() {
		// First interchange blocks an even number of times
		// Must be at least 3x2 or 2x2x2
		if ((blocks.x > 2 && (blocks.y > 1 || blocks.z > 1)) ||
			(blocks.y > 2 && (blocks.x > 1 || blocks.z > 1)) ||
			(blocks.z > 2 && (blocks.x > 1 || blocks.y > 1)) ||
			(blocks.x == 2 && blocks.y == 2 && blocks.z == 2)) {
			int	pos, count = 0;

			for (pos = 0; pos < sizeBlock; pos++) {
				int	randomPos = pos;

				while (randomPos == pos)
					randomPos = generator.nextInt(sizeBlock);
				if (exchangeBlocks(pos, randomPos))
					count++;
			}
			if ((count & 1) == 1 &&
					!exchangeBlocks(0, 1) &&
					!exchangeBlocks(sizeBlock - 2, sizeBlock - 1))
				System.out.println("randomizeBlocks: should not get here");
			allBlocksPaint = true;
			repaint();
		}
		// Now move the space around randomly
		if (blocks.x > 1 || blocks.y > 1 || blocks.z > 1) {
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_RESET);
			lastDirection = -1;
			randomPaints = java.lang.Math.min(sizeBlock,
				MAX_ITERATIONS) + generator.nextInt(2);
			repaint();
		}
	}

	void randomizingBlocks() {
		currentDirection = generator.nextInt(COORD);
		if (((currentDirection < 4 &&
				(currentDirection + 2) % 4 != lastDirection) ||
				(currentDirection >= 4 && currentDirection +
				(((currentDirection & 1) == 1) ? -1 : 1) !=
				lastDirection) ||
				(blocks.x == 1 && blocks.y == 1) ||
				(blocks.y == 1 && blocks.z == 1) ||
				(blocks.z == 1 && blocks.x == 1)) &&
				checkMoveBlocksDir(currentDirection)) {
			moveBlocks(currentPosition, INSTANT);
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_RANDOMIZE);
			started = false;
			currentPosition = -1;
			if (checkSolved())
				((CubesFrame) frame).callback(
					CubesInterface.ACTION_SOLVED);
		}
	}

	public boolean getNumber() {
		return number;
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		numSlices = ((delay < MAX_SLICES) ? delay + 1 : MAX_SLICES);
	}

	public boolean getCheat() {
		return cheat;
	}

	public boolean setPuzzleSizeX(int newValue) {
		if (blocks.x != newValue && newValue >= CubesInterface.MIN_BLOCKS) {
			//int vector = newValue - blocks.x;

			blocks.x = newValue;
			resetBlocks(/*vector*/);
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public boolean setPuzzleSizeY(int newValue) {
		if (blocks.y != newValue && newValue >= CubesInterface.MIN_BLOCKS) {
			//int vector = newValue - blocks.y;

			blocks.y = newValue;
			resetBlocks(/*vector*/);
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public boolean setPuzzleSizeZ(int newValue) {
		if (blocks.z != newValue && newValue >= CubesInterface.MIN_BLOCKS) {
			//int vector = newValue - blocks.z;

			blocks.z = newValue;
			resetBlocks(/*vector*/);
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((CubesFrame) frame).callback(CubesInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((CubesFrame) frame).callback(CubesInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((CubesFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		/* Determine size of client area */
		coreSize = new Point(getSize().width, getSize().height);
		delta = new Point(1, 1);
		offset = new Point(
			java.lang.Math.max((coreSize.x - 4) / blocks.x, 0),
			java.lang.Math.max((coreSize.y - 4) / blocks.y, 0));
		if (offset.y >= offset.x) {
			vertical = true;
			offset.y = java.lang.Math.max(((coreSize.y -
				5 * blocks.z + 1) / blocks.z) / blocks.y, 0);
		} else {
			vertical = false;
			offset.x = java.lang.Math.max(((coreSize.x -
				5 * blocks.z + 1) / blocks.z) / blocks.x, 0);
		}
		faceSize = new Point(offset.x * blocks.x + 4,
			offset.y * blocks.y + 4);
		if (vertical) {
			puzzleSize = new Point(faceSize.x,
				faceSize.y * blocks.z + blocks.z - 1);
		} else {
			puzzleSize = new Point(faceSize.x * blocks.z + blocks.z - 1,
				faceSize.y);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		blockSize = new Point(java.lang.Math.max(offset.x - delta.x,
			0), java.lang.Math.max(offset.y - delta.y, 0));
		bufferedFrameImage = createImage(coreSize.x, coreSize.y);
		bufferedFrameGraphics = bufferedFrameImage.getGraphics();
		if (number) {
			bufferedStartImage[0] =
				createImage(coreSize.x, coreSize.y);
			bufferedStartImage[1] =
				createImage(coreSize.x, coreSize.y);
		} else {
			bufferedStartImage[0] =
				createImage(PICTUREX, PICTUREY);
			bufferedStartImage[1] =
				createImage(PICTUREX, PICTUREY);
		}
		bufferedStartGraphics[0] = bufferedStartImage[0].getGraphics();
		bufferedStartGraphics[1] = bufferedStartImage[1].getGraphics();
	}

	public void initializePuzzle() {
		checkBlocks();
		undo = new CubesStack();
		redo = new CubesStack();
		solve = new CubesSolve(frame, this);
		resetBlocks(/*0*/);
		generator = new Random(System.nanoTime());
	}

	void exposePuzzle() {
		drawFrame(bufferedFrameGraphics, false);
		drawFrame(bufferedFrameGraphics, true);
		drawBuffer(bufferedStartGraphics[0], bufferedFrameImage);
		drawBuffer(bufferedStartGraphics[1], bufferedFrameImage);
		drawAllBufferedBlocks();
		drawBuffer(bufferedFrameImage);
		drawAllBlocks();
		solve.init();
	}

	void selectPuzzle(int x, int y) {
		currentPosition = positionToBlock(x, y);
		if (currentPosition >= 0) {
			if (checkSolved()) {
				moveNoBlocks();
				currentPosition = -1;
				return;
			}
			currentRow[0] = toColumn(currentPosition);
			currentRow[1] = toRow(currentPosition);
			currentRow[2] = toStack(currentPosition);
			selectPaint = true;
			repaint();
		} else
			currentPosition = -1;
	}

	void releasePuzzle() {
		if (currentPosition == -1)
			return;
		currentSpace = 0;
		releasePaint = true;
		repaint();
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetBlocks(/*0*/);
		drawAllBlocks();
		allBlocksPaint = true;
		repaint();
		((CubesFrame) frame).callback(
			CubesInterface.ACTION_RESET);
	}

	public void clearWithQueryPuzzle() {
		if (!started)
			clearPuzzle();
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizeBlocks();
	}

	public void getPuzzle() {
		this.stringSave = ((CubesFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();

		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			String symbol = SYMBOL;
			String newline = NL;
			int i, width, height, depth, moves;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			width = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (width >= CubesInterface.MIN_BLOCKS) {
				for (i = blocks.x; i < width; i++) {
					((CubesFrame) frame).callback(
						CubesInterface.ACTION_INCX);
				}
				for (i = blocks.x; i > width; i--) {
					((CubesFrame) frame).callback(
					CubesInterface.ACTION_DECX);
				}
			} else {
				showMessage(source +
					" corrupted: blocks.x " + width +
					" should be at least " +
					CubesInterface.MIN_BLOCKS);
				return;
			}
			lineData = token[++line];
			height = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (height >= CubesInterface.MIN_BLOCKS) {
				for (i = blocks.y; i < height; i++) {
					((CubesFrame) frame).callback(
						CubesInterface.ACTION_INCY);
				}
				for (i = this.blocks.y; i > height; i--) {
					((CubesFrame) frame).callback(
						CubesInterface.ACTION_DECY);
				}
			} else {
				showMessage(source +
					" corrupted: blocks.y " + height +
					" should be at least " +
					CubesInterface.MIN_BLOCKS);
				return;
			}
			lineData = token[++line];
			depth = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (depth >= CubesInterface.MIN_BLOCKS) {
				for (i = blocks.z; i < depth; i++) {
					((CubesFrame) frame).callback(
						CubesInterface.ACTION_INCZ);
				}
				for (i = blocks.z; i > depth; i--) {
					((CubesFrame) frame).callback(
						CubesInterface.ACTION_DECZ);
				}
			} else {
				showMessage(source +
					" corrupted: blocks.z " + depth +
					" should be at least " +
					CubesInterface.MIN_BLOCKS);
				return;
			}
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_RESTORE);
			setStartPosition();
			allBlocksPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": blocks.x " + blocks.x +
				", blocks.y " + blocks.y + ", blocks.z " + blocks.z +
				", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("blocks.x").append(symbol).append(space)
			.append(blocks.x).append(newline);
		sb.append("blocks.y").append(symbol).append(space)
			.append(blocks.y).append(newline);
		sb.append("blocks.z").append(symbol).append(space)
			.append(blocks.z).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((CubesFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		int direction = undo.readMoveDirection();

		redo.setMove(direction);
		// move in opposite direction
		direction = (direction < 4) ? (direction + 2) % 4 :
			((direction == INWARDS) ? OUTWARDS : INWARDS);
		if (checkMoveBlocksDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_UNDO);
		}
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		int direction = redo.readMoveDirection();

		undo.setMove(direction);
		if (checkMoveBlocksDir(direction)) {
			currentDirection = direction;
			stackPaint = true;
			repaint();
			((CubesFrame) frame).callback(
				CubesInterface.ACTION_REDO);
		}
	}

	void solvePuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		if (blocks.x == 1 || blocks.y == 1 || blocks.z == 1) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solveSomeBlocks();
		} else {
			showMessage("Auto-solver: sorry, only implemented for x = 1 or y = 1 or z = 1.");
		}
	}

	public void numberPuzzle() {
		number = !number;
		resizePaint = true;
		repaint();
	}

	void printPositions() {
		int	i;

		for (i = 0; i < sizeBlock; i++) {
			int pos = blockOfPosition[i];
			if (pos < 100)
				System.out.print(" ");
			if (pos < 10)
				System.out.print(" ");
			System.out.print(pos);
			if (((i + 1) % blocks.x) == 0) {
				System.out.println();
				if (((i + 1) % sizeRect) == 0)
					System.out.println();
			} else
				System.out.print(" ");
		}
		System.out.println();
	}

	void getStartPosition() {
		startSpace = spacePosition;
		System.arraycopy(blockOfPosition, 0, startPosition, 0, sizeBlock);
	}

	void setStartPosition() {
		spacePosition = startSpace;
		System.arraycopy(startPosition, 0, blockOfPosition, 0, sizeBlock);
		// drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int n = 0;

			++line; //read startingPosition:
			for (int k = 0; k < blocks.z; k++) {
				for (int j = 0; j < blocks.y; j++) {
					++line;
					while (token.length > line && token[line].equals(""))
						++line;
					if (line >= token.length) {
						System.out.println("Short data in StartPosition");
						return -1;
					}
					lineData = token[line];
					String[] st = lineData.split(space);
					int offset = ("".equals(st[0])) ? 1 : 0;
					for (int i = 0; i < blocks.x; i++) {
						int num = Integer.parseInt(st[i + offset].trim());

						startPosition[n] = num;
						if (num == 0)
							startSpace = n;
						n++;
					}
				}
				//blank line read
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input in StartPosition");
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;

			++line; //read moves direction
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				currentDirection = Integer.parseInt(
					lineData.substring(lineData.indexOf(symbol) +
					2));
				if (checkMoveBlocksDir(currentDirection)) {
					scanPaint = true;
					paint(g);
				} else {
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					currentDirection = -1;
					return false;
				}
			}
			if (l < moves) {
				showMessage("Corrupt input (scanMoves), only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading (scanMoves)");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < sizeBlock; i++) {
			int pos = startPosition[i];
			if (pos < 100)
				sb.append(space);
			if (pos < 10)
				sb.append(space);
			sb.append(space).append(pos);
			if (((i + 1) % blocks.x) == 0) {
				sb.append(newline);
				if (((i + 1) % sizeRect) == 0)
					sb.append(newline);
			}
		}
	}

	void printMoves(StringBuilder sb) {
		int direction;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space).append("direction").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			redo.setMove(direction);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(newline);
			undo.setMove(direction);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allBlocksPaint &&
				!selectPaint && !releasePaint &&
				!movePaint && !scanPaint &&
				!stackPaint && randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			drawFrame(bufferedFrameGraphics, true);
			framePaint = false;
		}
		if (allBlocksPaint) {
			drawAllBlocks();
			allBlocksPaint = false;
		}
		if (selectPaint) {
			if (currentPosition != -1) {
			int rowType = movableBlock();
			if (rowType < 0) {
				drawBlock(currentPosition,
					rowType == SPACE,
					false, 1, 0, 0);
				try {
					Thread.sleep((long) 3 * delay);
				} catch (InterruptedException e) {
					// ignore
				}
				drawBlock(currentPosition,
					true, true, 1, 0, 0);
				if (rowType != SPACE)
					drawBlock(currentPosition,
						false, false, 0, 0, 0);
				((CubesFrame) frame).callback(rowType);
				currentPosition = -1;
				mouseDown = false;
			} else
				drawBlock(currentPosition,
					false, false, 1, 0, 0);
			}
			selectPaint = false;
		}
		if (releasePaint) {
			if (currentPosition != -1) {
				drawBlock(currentPosition,
					true, true, 1, 0, 0);
				drawBlock(currentPosition,
					false, false, 0, 0, 0);
				if (currentSpace != -1)
					selectBlocks();
				currentPosition = -1;
			}
			releasePaint = false;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			if (currentPosition != -1) {
				moveBlocks(currentPosition, fast);
				started = true;
				if (!stackPaint && (movePaint || scanPaint)) {
					((CubesFrame) frame).callback(
					  CubesInterface.ACTION_MOVED);
					undo.setMove(currentDirection);
					redo.flushMoves();
					if (checkSolved())
						((CubesFrame) frame).callback(
							CubesInterface.ACTION_SOLVED);
					movePaint = false;
					scanPaint = false;
				}
				currentDirection = -1;
				currentPosition = -1;
			}
			stackPaint = false;
		}
		if (randomPaints > 0) {
			randomizingBlocks();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
